#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for ``moke.core``
"""
import unittest
import moke.core
import os
from moke.util import run_app

class Test_util(unittest.TestCase):

    def test_path(self):
        assert moke.util.path

    def setUp(self):
        os.chdir("scripts")

    def tearDown(self):
        os.chdir("..")
        
    def test_run1(self):
        os.chdir("..")
        ret, out, err,cmd =  run_app("moke")
        assert ret == 1
        assert out == ""
        assert err
        assert cmd == "moke"
        os.chdir("scripts")
        
    def test_run2(self):
        ret, out, err, cmd = run_app("moke")
        assert ret==2
        assert out == ""
        assert "usage: mokefile.py" in err
        assert cmd == "moke"

    def test_grop1(self):
        ret, out, err, cmd = run_app("./grop.py")
        assert ret == 2
        assert out == ""
        assert "usage: grop.py" in err
        assert cmd == './grop.py'

    def test_grop2(self):
        ret, out, err, cmd = run_app('cat ../data/grop.inp | ./grop.py ".*\(\d{2}\).*"')
        assert out == "a line with a number (42)\n"
        assert ret == 0

    def test_mf1(self):
        ret, out, err, cmd =  run_app("moke fromdef_int")
        assert ret == 0

    def test_mf2(self):
        ret, out, err, cmd =  run_app("moke fromdef_float")
        assert ret == 0

    def test_mf3(self):
        ret, out, err, cmd =  run_app("echo 1 | moke fromdef_path_r")
        assert ret == 0
        
    def test_mf3(self):
        ret, out, err, cmd =  run_app("echo 1 | moke fromdef_path_w")
        assert ret == 0

    def test_mf4(self):
        ret, out, err, cmd =  run_app("moke fromdoc_none_int -i 10")
        assert ret == 0, err


        
        
if __name__ == "__main__":
    unittest.main()


"""
:mod:`moke`
===========

Moke is not like ``make``. See the documentation.

"""
from core import *
from util import *
# this is for chained pipes of moke scripts
import signal
signal.signal(signal.SIGPIPE, signal.SIG_DFL)
del signal



"""
:mod:`moke.core`
================

``moke`` is not like ``make``. The core package provides the ``task`` the
``task`` all in one decorator.

"""

__all__ = ["MokeError", "task", "stdin", "stdout", "stderr", "num", "doc",
           "INFO", "WARN", "ERROR"]
__version__ = "1.0.0"


import os
import sys
import inspect
import logging
from re import search
from itertools import izip_longest
from argparse import ArgumentParser, FileType
from types import TypeType
from types import FileType as BaseFileType

# do not remove, really
from logging import INFO, WARN, ERROR
from sys import stdin, stdout, stderr
from path import path


__deflog__ = {"ls":stderr, # log stream
              "ll":"info", # log level
              "lf":"tab", # log format
             }

__logforms__ = {"tab":"%(asctime)s\t%(message)s"
                }

nan = float("nan")
devnull = open(os.devnull, "wb")

# monkey patching now to have simpler code later
file_w = FileType("wb")
file_w.__dict__["__name__"] = "file_w"
file_r = FileType("rb")
file_r.__dict__["__name__"] = "file_r"
file_a = FileType("a+")
file_a.__dict__["__name__"] = "file_a"

class MokeError(Exception):
    """General moke exception.
    """
    pass

def num(x):
    """(internal) check if value is a number.
    """
    try:
        num = int(x)
    except ValueError:
        try:
            num = float(x)
        except:
            num = nan
    return num

def doc(docstring):
    """Decorator that attaches a docstring to a function.
    
      - docstring(``str``) some documentation
    
    """
    def docwrapper(func):
        func.__doc__ = docstring
        return func

    return docwrapper


class task(object):
    """Decorator that creates an argparse subcommand from a function.
    """
    
    __funcs = {}
    

    @staticmethod
    def _makelgr(args):
        logstream = args.pop("ls")
        loglevel = getattr(logging, args.pop("ll").upper())
        logformat = __logforms__[args.pop("lf")]
        lgr = logging.getLogger("moke")
        lgr.setLevel(loglevel)
        sh = logging.StreamHandler(stream=logstream)
        sh.setFormatter(logging.Formatter(logformat))
        lgr.addHandler(sh)
        return lgr

    @staticmethod
    def _parsetype(line):
        # parses (sometypeX) or (``sometypeX``),
        # where X is a number or ? + * (argparse nargs)
        try:
            sometype, nargs = search("\(`{0,2}([a-z_]*)([\?\+\*]{1}|[0-9]?)`{0,2}\)",
                                     line).groups()
        except AttributeError:
            return None
        # test if sometype is a known type
        try:
            sometype = eval(sometype)
        except NameError, SyntaxError:
            return None
        if not (issubclass(type(sometype), TypeType) or \
                isinstance(sometype, FileType) or sometype is num):
            return None
        # fix empty string
        nargs = nargs or None
        if nargs not in ('+', '?', '*', None):
            try:
                nargs = int(nargs)
            except ValueError:
                return None
        return sometype, nargs

    @classmethod
    def _parsearg(cls, doclines, arg):
        docline, argtype, nargs = "", str, None # defaults
        # the space is mandatory for rst compatibility
        arg_type = "- " + arg + "(" # with type
        arg_noty = "- " + arg + " " # without type
        for line in doclines:
            if (arg_type in line or arg_noty in line) and line.startswith("-"):
                try:
                    docline = line.replace("-", "", 1).replace(arg, "", 1).strip()
                    argtype, nargs = cls._parsetype(line)
                except TypeError:
                        pass
                break # only the first match
        return docline, argtype, nargs

    @classmethod
    def _funcparse(cls):
        # parses gathered functions
        main_parser = ArgumentParser()

        # global options
        main_parser.add_argument("-ls", type=file_a,
            default=__deflog__["ls"],
            help="(file_a) [default: %s] logging stream" % __deflog__["ls"].name)

        main_parser.add_argument("-ll", type=str,
            default=__deflog__["ll"], choices=("info", "warn", "error"),
            help="(str) [default: %s] logging level" % __deflog__["ll"])

        main_parser.add_argument("-lf", type=str,
            default=__deflog__["lf"], choices=("tab",),
            help="(str) [default: %s] logging format" % __deflog__["lf"] )

        # subcommand options
        sub_parsers = None
        for name, func in cls.__funcs.iteritems():
            args, varargs, varkwargs, defaults = inspect.getargspec(func)
            if varargs or varkwargs:
                raise MokeError("*args and **kwargs not supported in tasks")
            arglines = []
            doclines = []
            if func.__doc__ and func.__doc__.strip():
                for l in func.__doc__.splitlines():
                    l = l.strip()
                    if l.startswith("-"):
                        arglines.append(l)
                    elif l:
                        doclines.append(l)
            if name == "main":
                task_parser = main_parser
                task_parser.description = "\n".join(doclines)
            else:
                if not sub_parsers:
                    sub_parsers = main_parser.add_subparsers()
                task_parser = sub_parsers.add_parser(name, help="\n".join(doclines))
            
            task_parser.set_defaults(func = func)
            if defaults:
                ldef = len(defaults)
                required = args[:-ldef]
                optional = args[-ldef:]
            else:
                required = args
                optional = ()
                defaults = ()
            
            for arg, deft in izip_longest(optional, defaults):
                docline, argtype, nargs = cls._parsearg(arglines, arg)
                if deft is False:
                    if not docline:
                        docline = "(switch) [default: OFF]"
                    task_parser.add_argument("--" + arg, action="store_true",
                                             default=deft, help=docline)
                elif deft is True:
                    if not docline:
                        docline = "(switch) [default: ON]"
                    task_parser.add_argument("--" + arg, action="store_false",
                                                 default=deft, help=docline)
                else:
                    # function signature > docs string
                    if deft is not None:
                        argtype = type(deft)
                    if argtype is BaseFileType:
                        # stdin, stdout or an open file
                        argtype = eval("file_" + deft.mode[0])

                    # default = most types
                    argname = argtype.__name__
                    argrepr = getattr(deft, "name", str(deft))
                    argdoc = "(%s) [default: %s] " % (argname, \
                                                argrepr.replace("%", "%%"))
                    if docline:
                        # (XXX) [default: YYY] CONTENT, (XXX) CONTENT, CONTENT
                        argdoc += docline.split(")")[-1].split("]")[-1].strip()
                    task_parser.add_argument("-" + arg, type=argtype, \
                                        nargs=nargs, default=deft, help=argdoc)
            for arg in required:
                docline, argtype, nargs = cls._parsearg(arglines, arg)
                argdoc = "(%s) " % argtype.__name__
                if docline:
                    argdoc += docline.split(")")[-1].split("]")[-1].strip()
                else:
                    argdoc += "a required positional argument"
                task_parser.add_argument(arg, type=argtype, nargs=nargs, \
                                         help=argdoc)
        return main_parser

    @classmethod
    def _callfunc(cls, args):
        # get function and logger
        func = args.pop("func")
        lgr = cls._makelgr(args)
        # where are we?
        cwd = os.getcwd()
        mokefile = os.path.abspath(os.path.join(cwd, sys.argv[0]))
        # remember default and command line args
        names, _, _, defs = inspect.getargspec(func)
        defs = (defs or ()) # defs can be None
        diff = len(names) - len(defs)

        vals = []
        for i, name in enumerate(names):
            try:
                vals.append(args[name])
            except KeyError:
                vals.append(defs[i - diff])

        params = " ".join(\
                [str(tpl) for tpl in zip(names, vals)])

        # write logging header
        msgs = ("moke (version %s)" % __version__,
                "cwd: \"%s\"" % cwd,
                "mokefile: \"%s\"" % mokefile,
                "task: %s" % func.func_name,
                "params: %s" % params)
        for msg in msgs:
            lgr.info(msg)
        # final call, leaving moke
        return func(**args)

    @classmethod
    def _call(cls):
        # calls the selected function with the parsed arguments
        parser = cls._funcparse()
        try:
            args = dict(parser.parse_args().__dict__)
        except Exception, e:
            return 1
        retval = cls._callfunc(args) # modifies args
        return retval

    def __init__(self, func=None):
        # gather functions or call selected sub-command
        if inspect.isfunction(func):
            self.__name = func.__name__
            self.__funcs[self.__name] = func
        else:
            self._call()

    def __call__(self, *args, **kwargs):
        return self.__funcs[self.__name](*args, **kwargs)



"""
:mod:`moke.util`
================

Utility functions for creating and interacting with command-line applications.

"""
__all__ = ["chk", "chk_exist", "chk_exit", "chk_pth", "chk_type", "chks",
           "istrue", "inp_dir", "inp_file", "log", "log_run", "mkdir",
           "out_dir", "out_file", "out_pth", "path", "run_app", "sh",
           "tmp_file"]

import os
import sys
import logging
from path import path
from tempfile import NamedTemporaryFile
from subprocess import Popen, PIPE


def chk_exit(status, cargo=None, always=False):
    """
    Kills Python with a status and error message.

    Arguments:

      - status(``int``) Exist status.
      - cargo(``str``) Error message [default: ``None``].
      - always(``bool``) Exit even if status is ``0`` [default: ``True``].

    """
    if status or always:
        if cargo:
            log(cargo, logging.ERROR)
        sys.exit(status)
    return cargo

def log(msg, level=logging.INFO):
    """
    Universal logging function.
    
    Arguments:
    
      - msg(``str``) message to log.
      - level(``int``) logging level [default: INFO].
     
    """
    lgr = logging.getLogger("moke")
    lgr.log(level, msg)
    
def run_app(cmd, args=None, stdin=None, **kwargs):
    """
    A simple wrapper around ``subprocess.Popen`` executes a command-line
    application with the given arguments and returns a 3-tuple of
    ``(returncode, stdout, stderr)``.
    
    Arguments:
    
      - app (``str``) The name of the application to run.
      - args (``sequence``) [default: ``None``] A sequence of arguments to the
        program.
      - stdin (``file``) A file-like object open for reading.
    
    Additional keyworded arguments are passed to the ``subprocess.Popen``
    constructor.
    
    """
    kwargs['stdout'] = kwargs.get('stdout', None) or PIPE
    kwargs['stderr'] = kwargs.get('stderr', None) or PIPE
    if args is not None:
        args.insert(0, cmd)
        app = Popen(args=args, stdin=stdin, **kwargs)
    else:
        kwargs["shell"] = True
        app = Popen(cmd, stdin=stdin, **kwargs)
    stdout, stderr = app.communicate(stdin)
    code = app.wait()
    return (code, stdout, stderr, " ".join(args or (cmd,)))

def log_run(code_stdout_stderr_cmd, silent=False):
    code, stdout, stderr, cmd = code_stdout_stderr_cmd
    if not code:
        log("Success: %s" % cmd)
        if not silent:
            log(stdout)
    else:
        log("Failure (code: %s): %s" % (code, cmd))
        if not silent:
            log("Errors:\n%s" % stdout + stderr)
    return code

def sh(command, **kwargs):
    """
    Executes command-line in the shell. No ``stdin``, ``stdout`` and ``stderr``
    go to ``/dev/null`` and the command is executed via the default shell.
    Additional keyworded arguments are passed to ``Popen``.
    
    Arguments:
    
      - command(``str`` or sequence of ``str``) This command line will be executed.
        See the documentation for "args" in ``subprocess.Popen``.
    
    """
    NULL = open(os.devnull, 'w')
    app = Popen(command, stdin=None, stdout=NULL, stderr=NULL, \
                                        shell=True, **kwargs)
    app.communicate()
    code = app.wait()
    return code

def tmp_file(filepath=None, overwrite=False, suffix=None):
    """
    Creates a temporary file and returns a file handle. If "filepath" is given
    it will be created or overwritten if overwrite is ``True``, else a temporary
    file will be created. The name of the file is accessible from the "name"
    attribute of the file handle, the file ceases to exist if it is closed or
    eventually disappears if the handle garbage collected.
    
    Arguments:
    
      - filepath(``str``) [default: ``None``] Path to the file.
      - overwrite(``bool``) [default: ``False``] Should the destination be
        overwritten if it exists? (ignored if "filepath" is ``None``)
    
    """
    suffix = (suffix or "")
    if filepath:
        if os.path.exists(filepath) and not overwrite:
            raise ValueError('The path: %s exists!' % filepath)
        fh = open(filepath, 'w+b')
    else:
        fh = NamedTemporaryFile(suffix=suffix)
    return fh

def istrue(cond, msg):
    if cond:
        return 0, ""
    else:
        return 1, msg

def chk_pth(pth):
    if not isinstance(pth, path):
        return 1, "not a valid path: %s" % pth
    return 0, pth

def chk_exist(pth, should_exist):
    notpath, pth = chk_pth(pth)
    if notpath:
        return 1, pth
    if should_exist and not pth.exists():
        return 1, "missing path: %s" % pth
    elif not should_exist and pth.exists():
        return 1, "blocking path: %s" % pth
    return 0, pth

def chk_type(pth, should_dir):
    notexist, pth = chk_exist(pth, True)
    if notexist:
        return 1, pth
    if should_dir and not pth.isdir():
        return 1, "path is not a directory: %s" % pth
    elif not should_dir and pth.isdir():
        return 1, "path is not a file: %s" % pth
    return 0, pth

def inp_dir(pth):
    return chk_type(pth, True)

def inp_file(pth):
    return chk_type(pth, False)

def out_dir(pth):
    return chk_exist(pth, False)

def out_file(pth):
    return chk_exist(pth, False)

def out_pth(pth):
    t1, err1 = chk_exist(pth, True)
    if t1:
        return t1, err1
    t2, err2 = chk_type(pth, True)
    if t2:
        return t2, err2
    return 0, pth

def chks(checks):
    checks = tuple(checks)
    errors = [check[1] for check in checks if check[0]]
    if errors:
        sys.exit("\n".join(errors))
    return [check[1].abspath() for check in checks]

def chk(check):
    return chks((check,))[0]

def mkdir(path):
    """
    Creates a directory and is silent if it already exists.
    
    Arguments:
    
      - path(``path``) A path object.
    
    """
    try:
        path.mkdir_p()
    except OSError, _:
        path.makedirs_p()
